//***************************************************************************
// Ces applets viennent en complment du livre :
//
//  		VIE ARTIFICIELLE
//      O la biologie rencontre l'informatique
//		Illustr en Java
//            Jean-Philippe Rennard
//           	  Vuibert 2002
//===========================================================================
//  Boids
//    v:0.1 12/09/2002
//
//  This program implements Boids as a demonstration of emergent behaviour
//
//    This program is free software; you can redistribute it and/or
//    modify it under the terms of the GNU General Public License
//    as published by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//***************************************************************************
package boid;

/** Un Boid */
public class Cindividu {
  static final float VITESSE_MAX=30;

  /** Pointeur sur processus */
  private Cboid boids;
  /** Identificateur */
  private int id;
  /** Position en x et y */
  CposXY position;
  /** Vitesse en x et y */
  CposXY vitesse;

  /** Construit un individu  position alatoire */
  public Cindividu(Cboid b, int rang) {
    boids=b;
    id=rang;
    position=new CposXY();
    vitesse=new CposXY();
    creeAleat(rang);
  }

  /** Construit un individu  position et vitesse alatoires */
  public void creeAleat(int rang) {
    boolean valide=true;
    Cindividu ind;
    do {
      valide=true;
      position.valX=(float)(Math.random()*boids.boidVue.lLargeur);
      position.valY=(float)(Math.random()*boids.boidVue.lHauteur);
      for(int i=0;i<rang;i++) {
        ind=(Cindividu)boids.lesIndividus.elementAt(i);
        if(Math.abs(distX(ind))<boids.distanceMini ||
                        Math.abs(distY(ind))<boids.distanceMini)
          valide=false;
      }
    } while(!valide);
    vitesse.valX=(float)(Math.random()*VITESSE_MAX);
    vitesse.valY=(float)(Math.random()*VITESSE_MAX);
  }
  /** Calcul de la nouvelle position */
  public void calcPos() {
    CposXY cohesion=new CposXY();
    deltaCohesion(cohesion);
    CposXY evite=new CposXY();
    deltaEvite(evite);
    CposXY aligne=new CposXY();
    deltaAligne(aligne);
    CposXY vent=new CposXY();
    deltaVent(vent);
    CposXY objectif=new CposXY();
    deltaObjectif(objectif);
    if(boids.disperseGen<100) {
      evite.valX *= 3.0;
      evite.valY *= 3.0;
      cohesion.valX *= -1.0;
      cohesion.valY *= -1.0;
    }
    vitesse.valX += cohesion.valX+evite.valX+aligne.valX+vent.valX+objectif.valX;
    vitesse.valX=(vitesse.valX>=0 ? Math.min(vitesse.valX,VITESSE_MAX) : Math.max(vitesse.valX,-1*VITESSE_MAX));
    position.valX = (position.valX+boids.boidVue.lLargeur+vitesse.valX/4)%
                                                        boids.boidVue.lLargeur;
    vitesse.valY += cohesion.valY+evite.valY+aligne.valY+vent.valY+objectif.valY;
    if(vitesse.valY>=0)
      vitesse.valY=Math.min(vitesse.valY,VITESSE_MAX);
    else
      vitesse.valY=Math.max(vitesse.valY,-1*VITESSE_MAX);
    position.valY = (position.valY+boids.boidVue.lHauteur+vitesse.valY/4)%boids.boidVue.lHauteur;
  }
  /** Calcul de la rgle de cohsion */
  private void deltaCohesion(CposXY centre) {
    int cpt=0;
    for(int i=0;i<boids.taillePopulation;i++) {
      Cindividu ind=(Cindividu)(boids.lesIndividus.elementAt(i));
      if(id!=ind.id) {
        float dist=calcDist(ind);
        if(dist<boids.boidVue.lLargeur/2.0) {
          cpt++;
          centre.valX += distX(ind);
          centre.valY += distY(ind);
        }
      }
    }
    if(cpt>0) {
      centre.valX /= cpt;
      centre.valY /= cpt;
    }
    centre.valX /= boids.valCohesion;
    centre.valY /= boids.valCohesion;
  }
  /** Calcul de la rgle d'vitement */
  private void deltaEvite(CposXY evite) {
    for(int i=0;i<boids.taillePopulation;i++) {
      Cindividu ind=(Cindividu)(boids.lesIndividus.elementAt(i));
      if(id!=ind.id) {
        float dx=distX(ind);
        float dy=distY(ind);
        if(Math.abs(dx)<boids.valEvite)
          evite.valX -= dx;
        if(Math.abs(dy)<boids.valEvite)
          evite.valY -= dy;
      }
    }
  }
  /** Calcul de la rgle d'alignement */
  private void deltaAligne(CposXY aligne) {
    int cpt=0;
    for(int i=0;i<boids.taillePopulation;i++) {
      Cindividu ind=(Cindividu)(boids.lesIndividus.elementAt(i));
      if(id!=ind.id) {
        float dist=calcDist(ind);
        if(dist<boids.boidVue.lLargeur/2.0) {
          cpt++;
          aligne.valX += ind.vitesse.valX;
          aligne.valY += ind.vitesse.valY;
        }
      }
    }
    aligne.valX /= cpt;
    aligne.valY /= cpt;
    aligne.valX=(aligne.valX-vitesse.valX)/boids.valAligne;
    aligne.valY=(aligne.valY-vitesse.valY)/boids.valAligne;
  }
  /** Calcul de la drive vers l'objectif */
  private void deltaObjectif(CposXY objectif) {
    if(!boids.objExiste) return;
    if(Math.abs(position.valX-boids.posObjectif.valX)<=boids.boidVue.lLargeur/2.0)
      objectif.valX += boids.posObjectif.valX-position.valX;
    else {
      if(boids.posObjectif.valX>=position.valX)
        objectif.valX += -1.0*(position.valX+(boids.boidVue.lLargeur-boids.posObjectif.valX));
      else
        objectif.valX += boids.posObjectif.valX+(boids.boidVue.lLargeur-position.valX);
    }
    if(Math.abs(position.valY-boids.posObjectif.valY)<=boids.boidVue.lHauteur/2.0)
      objectif.valY += boids.posObjectif.valY-position.valY;
    else {
      if(boids.posObjectif.valY>=position.valY)
        objectif.valY += -1.0*(position.valY+(boids.boidVue.lHauteur-boids.posObjectif.valY));
      else
        objectif.valY += boids.posObjectif.valY+(boids.boidVue.lHauteur-position.valY);
    }
    objectif.valX /= boids.valObjectif;
    objectif.valY /= boids.valObjectif;
  }
  /** Calcul de la direction gnrale (vent) */
  private void deltaVent(CposXY vent) {
    vent.valX=(float)(boids.valVent/10.0);
    vent.valY=(float)(boids.valVent/10.0);
  }

  /** Calcul de la distance en X */
  private float distX(Cindividu b) {
    float d=0;
    if(Math.abs(position.valX-b.position.valX)<=boids.boidVue.lLargeur/2.0)
      d = b.position.valX-position.valX;
    else {
      if(b.position.valX>=position.valX)
        d =(float)( -1.0*(position.valX+(boids.boidVue.lLargeur-b.position.valX)));
      else
        d = (float)(b.position.valX+(boids.boidVue.lLargeur-position.valX));
    }
    return d;
  }
  /** Calcul de la distance en Y */
  private float distY(Cindividu b) {
    float d=0;
    if(Math.abs(position.valY-b.position.valY)<=boids.boidVue.lHauteur/2.0)
      d = b.position.valY-position.valY;
    else {
      if(b.position.valY>=position.valY)
        d = (float)(-1.0*(position.valY+(boids.boidVue.lHauteur-b.position.valY)));
      else
        d = (float)(b.position.valY+(boids.boidVue.lHauteur-position.valY));
    }
    return d;
  }

  /** calcul de la distance d'un individu (Pythagore) */
  private float calcDist(Cindividu b) {
    return (float)(Math.sqrt(Math.pow(distX(b),2.0)+Math.pow(distY(b),2.0)));
  }
}


